# encoding: UTF-8
#--
# Copyright (c) 2014 Michael Berkovich, TranslationExchange.com
#
#  _______                  _       _   _             ______          _
# |__   __|                | |     | | (_)           |  ____|        | |
#    | |_ __ __ _ _ __  ___| | __ _| |_ _  ___  _ __ | |__  __  _____| |__   __ _ _ __   __ _  ___
#    | | '__/ _` | '_ \/ __| |/ _` | __| |/ _ \| '_ \|  __| \ \/ / __| '_ \ / _` | '_ \ / _` |/ _ \
#    | | | | (_| | | | \__ \ | (_| | |_| | (_) | | | | |____ >  < (__| | | | (_| | | | | (_| |  __/
#    |_|_|  \__,_|_| |_|___/_|\__,_|\__|_|\___/|_| |_|______/_/\_\___|_| |_|\__,_|_| |_|\__, |\___|
#                                                                                        __/ |
#                                                                                       |___/
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#++

class Tr8n::Language < Tr8n::Base
  belongs_to  :application
  attributes  :locale, :name, :english_name, :native_name, :right_to_left, :flag_url
  has_many    :contexts, :cases

  def self.cache_key(locale)
    File.join(locale, 'language')
  end

  def fetch
    update_attributes(application.api_client.get("language/#{locale}", {}, {:cache_key => self.class.cache_key(locale)}))
  rescue Tr8n::Exception => ex
    Tr8n.logger.error("Failed to load language: #{ex}")
    self
  end

  def update_attributes(attrs)
    super

    self.attributes[:contexts] = {}
    if hash_value(attrs, :contexts)
      hash_value(attrs, :contexts).each do |key, context|
        self.attributes[:contexts][key] = Tr8n::LanguageContext.new(context.merge(:keyword => key, :language => self))
      end
    end

    self.attributes[:cases] = {}
    if hash_value(attrs, :cases)
      hash_value(attrs, :cases).each do |key, lcase|
        self.attributes[:cases][key] = Tr8n::LanguageCase.new(lcase.merge(:keyword => key, :language => self))
      end
    end
  end

  def context_by_keyword(keyword)
    hash_value(contexts, keyword)
  end

  def context_by_token_name(token_name)
    contexts.values.detect{|ctx| ctx.applies_to_token?(token_name)}
  end

  def case_by_keyword(keyword)
    cases[keyword]
  end

  def has_definition?
    contexts.any?
  end

  def default?
    return true unless application
    application.default_locale == locale
  end

  def dir
    right_to_left? ? 'rtl' : 'ltr'
  end

  def align(dest)
    return dest unless right_to_left?
    dest.to_s == 'left' ? 'right' : 'left'
  end

  def full_name
    return english_name if english_name == native_name
    "#{english_name} - #{native_name}"
  end

  def current_source(options)
    (options[:source] || Tr8n.session.block_options[:source] || Tr8n.session.current_source || 'undefined').to_s
  end

  #######################################################################################################
  # Translation Methods
  #
  # Note - when inline translation mode is enable, cache will not be used and translators will
  # always hit the live service to get the most recent translations
  #
  # Some cache adapters cache by source, others by key. Some are read-only, some are built on the fly.
  #
  # There are three ways to call the tr method
  #
  # tr(label, description = "", tokens = {}, options = {})
  # or
  # tr(label, tokens = {}, options = {})
  # or
  # tr(:label => label, :description => "", :tokens => {}, :options => {})
  ########################################################################################################

  def translate(label, description = nil, tokens = {}, options = {})
    params = Tr8n::Utils.normalize_tr_params(label, description, tokens, options)
    return params[:label] if params[:label].tr8n_translated?

    translation_key = Tr8n::TranslationKey.new({
      :application  => application,
      :label        => params[:label],
      :description  => params[:description],
      :locale       => hash_value(params[:options], :locale) || hash_value(Tr8n.session.block_options, :locale) || Tr8n.config.default_locale,
      :level        => hash_value(params[:options], :level) || hash_value(Tr8n.session.block_options, :level) || Tr8n.config.default_level,
      :translations => []
    })

    #Tr8n.logger.info("Translating " + params[:label] + " from: " + translation_key.locale.inspect + " to " + locale.inspect)

    params[:tokens] ||= {}
    params[:tokens][:viewing_user] ||= Tr8n.session.current_user

    if Tr8n.config.disabled? or self.locale == translation_key.locale
      return translation_key.substitute_tokens(
        params[:label],
        params[:tokens],
        self,
        params[:options]
      ).tr8n_translated
    end

    cached_translations = application.cached_translations(locale, translation_key.key)
    if cached_translations
      translation_key.set_translations(locale, cached_translations)
      return translation_key.translate(self, params[:tokens], params[:options]).tr8n_translated
    end

    source_key = current_source(options)

    if Tr8n.cache.segmented? or Tr8n.session.inline_mode?
      source = application.source(source_key, locale)
      cached_translations = source.cached_translations(locale, translation_key.key)
    else
      cached_translations = application.fetch_translations(locale)[translation_key.key]
    end

    if cached_translations
      translation_key.set_translations(locale, cached_translations)
    else
      application.register_missing_key(source_key, translation_key)
    end

    translation_key.translate(self, params[:tokens], params[:options]).tr8n_translated
  end
  alias :tr :translate

end
